/**
 *
 */

#include <iostream>

#include <asio.hpp>
#include <blt/std/logging.h>
#include <blt/std/memory.h>
#include <string>
#include <array>
#include <thread>
#include <iomanip> // This might be necessary
#include <vector>
#include <type_traits>

using asio::ip::udp;
using asio::ip::tcp;

// DNS data contains:
// 2 bytes for transaction id
// 2 bytes for flags
// 2 bytes for number of questions
// 2 bytes for Answer RRs
// 2 bytes for Authority RRs
// 2 bytes for Additional RRs

// question format:
// 1 byte for length
// (length) bytes per section
// ... until length 0
// 2 bytes for QTYPE (useless)
// 2 bytes for QCLASS (useless)

// answer format:
// 2 byte for domain name (offset ptr, still not 100 on from where)
// 2 byte for type
// 2 byte for class
// 4 byte for time to live
// 2 byte for length of data
// (lengthy) byte for data

class send_buffer
{
    private:
        mutable std::array<unsigned char, 65535> internal_data{};
        mutable size_t write_index = 0;
    public:
        send_buffer() = default;
        
        template<typename T>
        void write(const T& t) const
        {
            if constexpr (std::is_arithmetic_v<T>)
            {
                blt::mem::toBytes(t, &internal_data[write_index]);
                write_index += sizeof(T);
            } else if constexpr (std::is_same_v<T, std::string>)
            {
                std::memcpy(&t[0], t.size());
                write('\0');
            } else
                static_assert("Data type not supported!");
        }
        
        void reset() const
        {
            write_index = 0;
        }
        
        [[nodiscard]] unsigned char operator[](size_t i){
            return internal_data[i];
        }
        
        [[nodiscard]] unsigned char* data()
        {
            return internal_data.data();
        }
        
        [[nodiscard]] size_t size() const
        {
            return write_index;
        }
        
        auto buffer() const
        {
            return asio::buffer(internal_data.data(), size());
        }
};

class byte_reader
{
    private:
        unsigned char* data;
        mutable size_t current_byte = 0;
    public:
        explicit byte_reader(unsigned char* data): data(data)
        {
            reset();
        }
        
        inline void reset() const
        {
            // magic number for end of header
            current_byte = 12;
        }
        
        inline unsigned char& next() const
        {
            return data[current_byte++];
        }
        
        template<typename T>
        inline void to(T& t) const
        {
            blt::mem::fromBytes(&data[current_byte], t);
            skip(sizeof(T));
        }
        
        inline void copy(unsigned char*& out, size_t size) const
        {
            std::memcpy(out, &data[current_byte], size);
            current_byte += size;
        }
        
        inline void skip(size_t s = 1) const
        {
            current_byte += s;
        }
};

class question
{
    private:
        std::string domain;
    public:
        explicit question(const byte_reader& reader)
        {
            // process the full question
            while (true)
            {
                uint8_t length = reader.next();
                if (length == 0)
                    break;
                if (!domain.empty())
                    domain += '.';
                for (uint8_t j = 0; j < length; j++)
                    domain += static_cast<char>(reader.next());
            }
            // Skip QTYPE and QCLASS
            reader.skip(4);
        }
        
        const std::string& operator()()
        {
            return domain;
        }
};

class answer
{
    private:
        uint16_t NAME = 0;
        uint16_t TYPE = 0;
        uint16_t CLASS = 0;
        uint32_t TTL = 0;
        uint16_t RDLENGTH = 0;
        unsigned char* RDATA = nullptr;
    public:
        explicit answer(const byte_reader& reader)
        {
            reader.to(NAME);
            reader.to(TYPE);
            reader.to(CLASS);
            reader.to(TTL);
            reader.to(RDLENGTH);
            RDATA = new unsigned char[RDLENGTH];
            reader.copy(RDATA, RDLENGTH);
            
            BLT_TRACE("%d, %d, %d, %d, %d", NAME, TYPE, CLASS, TTL, RDLENGTH);
        }
        
        // rule of 3
        answer(const answer& answer) = delete;
        
        answer& operator=(const answer& answer) = delete;
        
        ~answer()
        {
            delete[] RDATA;
        }
};

template<typename IN, typename OUT>
void sendUDPMessage(const std::string& host, const IN& in, size_t in_size, OUT& out, size_t& out_size)
{
    asio::io_context io_context;
    udp::resolver resolver(io_context);
    udp::endpoint receiver_endpoint(asio::ip::address::from_string(host), 53);
    
    udp::socket socket(io_context);
    socket.open(udp::v4());
    
    socket.send_to(asio::const_buffers_1(in, in_size), receiver_endpoint);
    
    udp::endpoint sender_endpoint;
    out_size = socket.receive_from(asio::buffer(out), sender_endpoint);
}

struct udp_server
{
        explicit udp_server(asio::ip::udp::socket socket)
                : local_socket(std::move(socket))
        {
            read();
        }
    
    private:
        void handle_receive(asio::ip::udp::endpoint ep, std::size_t length)
        {
            // get the number of questions
            uint16_t questions; // yes I made this part of my library just for this :3
            blt::mem::fromBytes(&recv_buf.data()[4], questions); // i hate little endian

            // the reader starts after the header to make things easier
            byte_reader reader(recv_buf.data());

            BLT_INFO("Bytes received: %d with %d questions", length, questions);

            // no one actually does multiple questions. trying to do it in dig is not easy
            // and the standard isn't really designed for this (how do we handle if one question errors but the other doesn't? there is only
            // one return code.)
            for (uint16_t i = 0; i < questions; i++)
            {
                question q(reader);
                BLT_INFO("DOMAIN: %s", q().c_str());
            }
            
            //write(ep);
        }
        
        void handle_write()
        {
        
        }
        
        void read()
        {
            local_socket.async_receive_from(asio::buffer(recv_buf), remote_connection, [this](asio::error_code ec, std::size_t length) {
                                                if (ec)
                                                    return;
                                                this->handle_receive(remote_connection, length);
                                                this->read();
                                            }
            );
        }
        
        void write(const asio::ip::udp::endpoint& remote)
        {
            std::thread th([&, this]() -> void {
                handle_write();
                local_socket.async_send_to(send_buf.buffer(), remote, [this](asio::error_code ec, std::size_t length) {
                                               if (ec)
                                                   return;
                                               send_buf.reset();
                                               this->read();
                                           }
                );
            });
        }
        
        asio::ip::udp::socket local_socket;
        asio::ip::udp::endpoint remote_connection;
        std::array<unsigned char, 65565> recv_buf{};
        send_buffer send_buf{};
};


int main()
{
//    try
//    {
//        asio::io_context io_context;
//
//        udp::socket socket(io_context, udp::endpoint(udp::v6(), 5555));
//
//        while (true)
//        {
//            std::array<unsigned char, 65535> recv_buf{};
//            std::array<unsigned char, 65535> mod_recv_buf{};
//            udp::endpoint remote_endpoint;
//            size_t bytes = socket.receive_from(asio::buffer(recv_buf), remote_endpoint);
//
//            // get the number of questions
//            uint16_t questions; // yes I made this part of my library just for this :3
//            blt::mem::fromBytes(&recv_buf[4], questions); // i hate little endian
//
//            // the reader starts after the header to make things easier
//            byte_reader reader(recv_buf.data());
//
//            BLT_INFO("Bytes received: %d with %d questions", bytes, questions);
//
//            // no one actually does multiple questions. trying to do it in dig is not easy
//            // and the standard isn't really designed for this (how do we handle if one question errors but the other doesn't? there is only
//            // one return code.)
//            for (uint16_t i = 0; i < questions; i++)
//            {
//                question q(reader);
//                BLT_INFO("DOMAIN: %s", q().c_str());
//            }
//
//            size_t out_bytes;
//            sendUDPMessage("8.8.8.8", recv_buf.data(), bytes, mod_recv_buf, out_bytes);
//
//            byte_reader reader2(mod_recv_buf.data());
//            question q(reader2);
//            answer a(reader2);
//
//            asio::error_code ignored_error;
//            socket.send_to(asio::const_buffers_1(mod_recv_buf.data(), out_bytes), remote_endpoint, 0, ignored_error);
//        }
//    }
//    catch (std::exception& e)
//    {
//        std::cerr << e.what() << std::endl;
//    }
    
    try
    {
        asio::io_context io_context;

        asio::ip::udp::endpoint ep(asio::ip::udp::v6(), 5555); // also listens on ipv4
        asio::ip::udp::socket sock(io_context, ep);
        udp_server server(std::move(sock));
        io_context.run();
    }
    catch (std::exception& e)
    {
        std::cerr << e.what() << std::endl;
    }
    
    std::cout << "Hello, World!" << std::endl;
    return 0;
}
